/* $Id: playsample.c,v 1.4 1994/09/23 00:27:06 jbyrd Exp $ */
/***************************************************************
**
** Play AIFF Sample in memory using Control Pad
** By:  Phil Burk
**
** Link with the music.lib to get SelectSamplePlayer()
**
** Copyright (c) 1992, 3DO Company.
** This program is proprietary and confidential.
***************************************************************/
/**
|||	AUTODOC PUBLIC sample/playsample
|||	playsample  - alpha autodoc header..summary to be added later
|||
|||	  Synopsis
|||
|||	    playsample [sampleFileName] [rate]
|||
|||	  Description
|||
|||	    This program shows how to load an AIFF sample file and
|||	    play it using the control pad.  Use the A button to
|||	    start the sample, the B button to release the sample,
|||	    and the C button to stop the sample.  The X button
|||	    quits the program.
|||
|||	  Arguments
|||
|||	    sampleFileName	The path and file name of a
|||				compatible AIFF sample file.
|||
|||	    rate		The rate at which to play back
|||				the sample, in 16.16 Hz.
|||
|||	  Caveats
|||
|||	    None.
|||
**/

#include "types.h"
#include "filefunctions.h"
#include "debug.h"
#include "operror.h"
#include "stdio.h"
#include "stdlib.h"
#include "event.h"

/* Include this when using the Audio Folio */
#include "audio.h"
/* Include this when using the music.lib */
#include "music.h"

/* Handy printing and debugging macros. */
#define	PRT(x)	{ printf x; }
#define	ERR(x)	PRT(x)
#define	DBUG(x)	/* PRT(x) */

/* Macro to simplify error checking. */
#define CHECKRESULT(val,name) \
	if (val < 0) \
	{ \
		Result = val; \
		PrintError(0,"\\failure in",name,val); \
		goto cleanup; \
	}

/************************************************************/
int main(int argc, char *argv[])
{
	Item OutputIns = 0;
	Item SamplerIns = 0;
	Item SampleItem = 0;
	Item Attachment = 0;
	char *SampleName = "sinewave.aiff";
	int32 Rate = 0x8000;
	int32 Result = -1;
	int32 DoIt = TRUE;
	int32 IfVariable = FALSE;
	char *InstrumentName;
	uint32 Buttons;
	ControlPadEventData cped;
	TagArg SamplerTags[3];
	
	PRT(("Usage: %s <samplefile> <rate>\n", argv[0]));

/* Get optional argumants from command line. */
	if (argc > 1) SampleName = argv[1]; 
	if (argc > 2)
	{
		Rate=atoi(argv[2]);
		IfVariable = TRUE;
	}
	
/* Print menu of button commands. */
	PRT(("Button Menu:\n"));
	PRT(("   A = Start\n"));
	PRT(("   B = Release\n"));
	PRT(("   C = Stop\n"));
	PRT(("   X = Exit\n"));
	
/* Initialize audio, return if error. */ 
	Result = OpenAudioFolio();
	if (Result < 0)
	{
		PrintError(0,"Audio Folio could not be opened.",0,Result);
		return(-1);
	}

/* Initialize the EventBroker. */
	Result = InitEventUtility(1, 0, LC_ISFOCUSED);
	if (Result < 0)
	{
		PrintError(0,"InitEventUtility",0,Result);
		goto cleanup;
	}

/* Load a directout instrument to send the sound to the DAC. */
	OutputIns = LoadInstrument("directout.dsp",  0, 0);
	CHECKRESULT(OutputIns,"LoadInstrument");
	StartInstrument( OutputIns, NULL );

/* Load digital audio AIFF Sample file from disk. */
	SampleItem = LoadSample(SampleName);
	CHECKRESULT(SampleItem,"LoadSample");
	
/* Look at sample information for fun. */
	DebugSample(SampleItem);
	
/* Select an apropriate sample playing instrument based on sample format. */
	InstrumentName = SelectSamplePlayer( SampleItem , IfVariable );
	if (InstrumentName == NULL)
	{
		ERR(("No instrument to play that sample.\n"));
		goto cleanup;
	}

/* Load Sampler instrument */
	PRT(("Use instrument: %s\n", InstrumentName));
	SamplerIns = LoadInstrument(InstrumentName,  0, 100);
	CHECKRESULT(SamplerIns,"LoadInstrument");

/* Connect Sampler Instrument to DirectOut. Works for mono or stereo. */
	Result = ConnectInstruments (SamplerIns, "Output", OutputIns, "InputLeft");
	if( Result >= 0 )
	{
		Result = ConnectInstruments (SamplerIns, "Output", OutputIns, "InputRight");
		CHECKRESULT(Result,"ConnectInstruments");
	}
	else
	{
		Result = ConnectInstruments (SamplerIns, "LeftOutput", OutputIns, "InputLeft");
		CHECKRESULT(Result,"ConnectInstruments");
		Result = ConnectInstruments (SamplerIns, "RightOutput", OutputIns, "InputRight");
		CHECKRESULT(Result,"ConnectInstruments");
	}
	
/* Attach the sample to the instrument for playback. */
	Attachment = AttachSample(SamplerIns, SampleItem, 0);
	CHECKRESULT(Attachment,"AttachSample");

/*
** Start the instrument at the given "frequency".  You could also
** use AF_TAG_PITCH, AF_TAG_AMPLITUDE.
*/
	SamplerTags[0].ta_Tag = AF_TAG_FREQUENCY;
	SamplerTags[0].ta_Arg = (void *) Rate;
	SamplerTags[1].ta_Tag = TAG_END;
	Result = StartInstrument( SamplerIns, SamplerTags );
	
/* Interactive event loop. */
	while(DoIt)
	{
/* Get User input. */
		Result = GetControlPad (1, TRUE, &cped);
		if (Result < 0) {
			PrintError(0,"read control pad in","PlaySoundFile",Result);
		}
		Buttons = cped.cped_ButtonBits;
	
/* Process buttons pressed. */
		if(Buttons & ControlX) /* EXIT */
		{
			DoIt = FALSE;
		}
		if(Buttons & ControlA) /* START */
		{
			Result = StartInstrument( SamplerIns, SamplerTags );
			CHECKRESULT(Result,"StartInstrument");
		}
		if(Buttons & ControlB) /* RELEASE */
		{
			Result = ReleaseInstrument( SamplerIns, NULL );
			CHECKRESULT(Result,"ReleaseInstrument");
		}
		if(Buttons & ControlC) /* STOP */
		{
			Result = StopInstrument( SamplerIns, NULL );
			CHECKRESULT(Result,"StopInstrument");
		}
	}

	Result = StopInstrument( SamplerIns, NULL );
	CHECKRESULT(Result,"StopInstrument");
	
cleanup:

	DetachSample( Attachment );
	UnloadSample( SampleItem );
	UnloadInstrument( SamplerIns );
	UnloadInstrument( OutputIns );
	
/* Cleanup the EventBroker. */
	Result = KillEventUtility();
	if (Result < 0)
	{
		PrintError(0,"KillEventUtility",0,Result);
	}
	CloseAudioFolio();
	PRT(( "%s finished.\n", argv[0] ));
	return((int) Result);
}
